/*
** Module   : ResourceIdHelper.java
** Abstract : Facility to encapsulate semantics for Progress resource identifiers. 
**
** Copyright (c) 2013-2014, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 HC  20131030 Created initial version, see issue #2183.
** 002 CA  20131219 Negative resource IDs must not be generated, as they are not valid resource 
**                  IDs in 4GL (thus a idFromString call may fail if resource ID is negative, even 
**                  if it is a real ID).
** 003 GES 20140115 Documentation changes, added comments and added a small feature to
**                  idFromString().
*/

package com.goldencode.p2j.util;

import java.math.BigInteger;
import java.security.SecureRandom;


/** 
 * The class encapsulates semantics related to Progress resource identifiers, 
 * see {@link handle#resourceId(WrappedResource)}.
 * <p>
 * The main areas of functionality implemented by this class include allocations
 * of new resource identifiers, string formatting and parsing.
 * <p>
 * A resource identifier takes two forms, numeric and textual. The numeric form is 
 * represented by {@link Long} data type.
 * <p>
 * Internally, P2J can avoid the string representation of the resource ID and use it as a long
 * value. For this reason, the {@link #getNext()} API will never generate negative resource IDs, 
 * as these are not valid legacy resource IDs (as calling {@link #idFromString} with a negative
 * resource ID will fail).
 */
public class ResourceIdHelper
{
   /** Constant denoting the unknown resource id */
   public static final Long UNKNOWN = null;
   
   /** Constant denoting the unknown resource id string */
   public static final String UNKNOWN_STRING = "?";
   
   /** Constant denoting the zero resource id */
   public static final Long ZERO = 0l;
   
   /** The default format string for a resource identifier (WARNING: not compatible with Progress). */
   public static final String DEFAULT_FORMAT = ">>>>>>>>>>>>>>>>>>>9";   
   
   /** 
    * We are using Secure random number generator so the generated values
    * are uniformly distributed in the whole 64-bit range.
    */
   private static SecureRandom rnd = new SecureRandom();
   
   /** Max long constant */
   private static final BigInteger MAX_LONG = BigInteger.valueOf(Long.MAX_VALUE);
   
   /**
    * Generates a valid resource identifier.
    * <p> 
    * The returned value is a randomly generated positive 64bit number with no guarantee of 
    * uniqueness. It is the responsibility of the caller to check uniqueness of the returned 
    * value in the caller's context if required.
    */
   public static Long getNext() 
   {
      long num = 0;
      
      // zero doesn't denote a valid id
      while (num == 0) 
      {
         num = rnd.nextLong();
      }
      
      // do not generae negative resource IDs.
      if (num < 0)
      {
         num = num + Long.MAX_VALUE;
      }
      
      return num;
   }
   
   /**
    * Returns <code>true</code> if the supplied identifier represents
    * a zero resource identifier, <code>false</code> otherwise.
    * 
    * @param   resourceId
    *          Resource identifier.
    */
   public static boolean isZero(Long resourceId) 
   {
      return ZERO.equals(resourceId);
   }
   
   /**
    * Returns <code>true</code> if the supplied identifier represents
    * an unknown resource identifier, <code>false</code> otherwise.
    * 
    * @param   resourceId
    *          Resource identifier.
    * 
    */
   public static boolean isUnknown(Long resourceId) 
   {
      return UNKNOWN == resourceId;
   }
   
   /**
    * Parses a resource id from its textual form into its numeric form. The number is parsed into
    * a 64-bit integer using Progress conventions, see {@link NumberType#parseDouble}.
    * <p>
    * The method returns {@code UNKNOWN} if the input is {@code null} or equal to "?" or
    * the method returns {@code ZERO} if the input contains a trimmed empty string or
    * the method returns {@code ZERO} if the parsed input is zero or less than zero or
    * the method throws {@link ErrorConditionException} if the input cannot be parsed to 
    * an integer.
    * 
    * @param    resourceId
    *           A string value representing a resource identifier.
    *
    * @return   A numeric resource id.
    *
    * @throws   ErrorConditionException
    *           If the input cannot be parsed into an integer, see method description 
    *           for more details.
    */
   public static Long idFromString(String resourceId)
   throws ErrorConditionException
   {
      if (resourceId == null || resourceId.equals(UNKNOWN_STRING))
      {
         return UNKNOWN;
      }
      
      // we don't know why this is a special case in the 4GL, but it is
      if (resourceId.length() > 255)
      {
         return ZERO;
      }
      
      // TODO: report two errors: "** Decimal number is too large. (536)" and 
      //                          "** Value too large for integer. (78)"
      //    when the string exceeds 50 significant digits (i.e. not including numsign, 
      //    white spaces, leading zeros, etc.).
      //    UPDATE: this would probably be obtained "for free" by utilizing a code path that
      //            runs through int64.setValue(String) which already has this same 2 error
      //            (536 then 78) behavior; parseDecimal() already does throw 536 and the
      //            int64.setValue() uses parseDecimal() and then if it gets a 536 it raises
      //            the 78 error; converted code tests should be run to confirm
      
      // normalize the string with Progress conventions
      String javaResourceId = NumberType.parseDecimal(resourceId, true);
      BigInteger num = new BigInteger(javaResourceId);
      
      // we only consider positive integers valid resource ids. this is safe because P2J will
      // never generate a negative resource ID
      if (num.signum() <= 0)
      {
         return ZERO;
      }
      
      // out of range of positive long?
      if (num.compareTo(MAX_LONG) > 0)
      {
         ErrorManager.recordOrThrowError(78, "Value too large for integer");
         return null;
      }
      
      return num.longValue();
   }
   
   /**
    * Formats the numeric resource identifier with the default format string,
    * {@link ResourceIdHelper#DEFAULT_FORMAT}.
    * See {@link #idToString(String, Long)} for more details.
    * 
    * @param   resourceId
    *          Resource identifier.
    * 
    * @return  the formatted resource identifier.
    */
   public static String idToString(Long resourceId) 
   {
      return idToString(null, resourceId);
   }
   
   /**
    * Formats the numeric resource identifier with the supplied format string.
    * If this method receives a negative resource ID, an exception is thrown. Internally, P2J will
    * never use negative IDs.
    * 
    * @param   format
    *          Format string. 
    * @param   resourceId
    *          Resource identifier.
    * 
    * @return  the formatted resource identifier.
    * 
    * @throws  IllegalArgumentException
    *          If the resource ID is negative.
    */
   public static String idToString(String format, Long resourceId) 
   {
      if (isUnknown(resourceId))
      {
         return UNKNOWN_STRING;
      }
      
      BigInteger num = BigInteger.valueOf(resourceId);
      
      if (num.signum() < 0) 
      {
         throw new IllegalArgumentException("The resource ID can not be negative!");
      }
      
      return NumberType.toString(num.toString(), "", false, format == null ? DEFAULT_FORMAT : format, false);
   }
}